function [sampleLabels, subspaceBases, returnStatus] = RANSACsubspaces(X, subspaceDimensions, varargin)

DEBUG = 1;

warning off
% Set default parameter values
maxIteration = 0;
confidencePercentage = 0;                       % If both maxIteration and confidencePercentage are assigned, prefer latter
maxOutlierPercentage = 0.75;                    % Used together with confidencePercentage
postOptimization = true;
boundaryThreshold = 0.1;
[ambientDimension sampleCount]= size(X);
subspaceCount = length(subspaceDimensions);
consensusThreshold = 1;
angleTolerance = 0;
dataNormalization = true;
degeneracyThreshold = 0.1;                      % This is a threshold value for degeneracy test.

% Parse the optional inputs.
if (mod(length(varargin), 2) ~= 0 ),
    error(['Extra Parameters passed to the function ''' mfilename ''' must be passed in pairs.']);
end
parameterCount = length(varargin)/2;

parameterInitialized = false;
for parameterIndex = 1:parameterCount,
    parameterName = varargin{parameterIndex*2 - 1};
    parameterValue = varargin{parameterIndex*2};
    switch lower(parameterName)
        case 'postoptimization'
            postOptimization = parameterValue;
            if postOptimization~=1 && postOptimization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end 
        case 'datanormalization'
            dataNormalization = parameterValue;
            if dataNormalization~=1 && dataNormalization~=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'maxiteration'
            maxIteration = parameterValue;
            if maxIteration<10
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'boundarythreshold'
            boundaryThreshold = parameterValue;
            if boundaryThreshold<=0
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        case 'confidencepercentage'
            confidencePercentage = parameterValue;
            if confidencePercentage<=0 || confidencePercentage>=1
                error(['The parameter ''' parameterName ''' is out of range']);
            end
        otherwise
            error(['Sorry, the parameter ''' parameterName ''' is not recognized by the function ''' mfilename '''.']);
    end
end

if (confidencePercentage==0) && (maxIteration==0)
    error('RANSAC stop criterion is not given');
end

% Normalize data vectors
sampleMagnitudes = zeros(1,sampleCount);
for sampleIndex=1:sampleCount
    sampleMagnitudes(sampleIndex)=norm(X(:,sampleIndex));
end
X = X/max(sampleMagnitudes);
if dataNormalization
    [X, subspaceVariances]=normalize_variance(X);
end

% Iteratively extract subspaces from high to low
[sortedDimensions, sortedIndex]= sort(subspaceDimensions, 'descend');

if confidencePercentage>0
    iterationNumber = max_RANSAC_iteration_with_confidence(confidencePercentage,maxOutlierPercentage,...
        subspaceCount-subspaceIndex+1, sortedDimensions(subspaceIndex));
else
    iterationNumber = maxIteration;
end

leftX = X; leftXCount = sampleCount;
for subspaceIndex=1:subspaceCount
    if DEBUG>0
        disp(['Estimating subspace #' num2str(subspaceIndex)]);
    end
    
    currentX = leftX;
    currentXCount = leftXCount;
    currentSubspaceDimension = sortedDimensions(subspaceIndex);
    degeneracyDimension = sortedDimensions(subspaceIndex)-1;
    if currentXCount<currentSubspaceDimension
        error('RANSAC failed to segment a generally distributed subspace model.');
    end
    
    maxSubspaceConsensus = 0;
    adjustedIterationNumber = nchoosek(currentXCount,currentSubspaceDimension);
    if iterationNumber>adjustedIterationNumber
        newSamplingIndices = generate_subsets(currentXCount,currentSubspaceDimension);
    else
        adjustedIterationNumber = iterationNumber;
        newSamplingIndices = [];
        for iterationIndex=1:iterationNumber
            newSamplingIndices(iterationIndex,:) = generate_random_subset(currentXCount,currentSubspaceDimension);
        end
    end
    
    for iterationIndex=1:adjustedIterationNumber
        % Step 1: generate a minimal sample set
        subsamples=currentX(:,newSamplingIndices(iterationIndex,:));

        % Step 2: Calculate an orthogonal basis
        candidateBasis = orth(subsamples);

        % Step 3: Calculate the consensus
        distance = point_to_space_distance(currentX,candidateBasis);
        outlierIndices = find(distance>boundaryThreshold);
        inlierIndices = find(distance<=boundaryThreshold);

        if length(inlierIndices)>maxSubspaceConsensus
            maxSubspaceConsensus = length(inlierIndices);
            maxOutlierIndices = outlierIndices;
            maxInlierIndices = inlierIndices;
        end
    end
    
    
    if sum(sortedDimensions(subspaceIndex:subspaceCount)~=currentSubspaceDimension)==0
        degeneracyExist = false;
        currentX = currentX(:,maxInlierIndices);
    else
        degeneracyExist = true;
        possibleIntersectionNumber = subspaceCount-subspaceIndex+1;
    end
    while degeneracyExist
        % Check for degeneracy
        currentInliers = currentX(:,maxInlierIndices);
        currentInlierCount = maxSubspaceConsensus;
        degeneracyExist = false;
        for degeneracyIndex=1:possibleIntersectionNumber
            maxDegeneracyConsensus = 0;
            maxDegeneracyOutlierIndices = [];
            if currentInlierCount<degeneracyDimension
                error('RANSAC failed to segment a generally distributed subspace model.');
            end
            adjustedIterationNumber = nchoosek(currentInlierCount,degeneracyDimension);
            if iterationNumber>adjustedIterationNumber
                newSamplingIndices = generate_subsets(currentInlierCount,degeneracyDimension);
            else
                adjustedIterationNumber = iterationNumber;
                newSamplingIndices = [];
                for iterationIndex=1:iterationNumber
                    newSamplingIndices(iterationIndex,:) = generate_random_subset(currentInlierCount,degeneracyDimension);
                end
            end

            for iterationIndex =1:adjustedIterationNumber
                subsamples = currentInliers(:,newSamplingIndices(iterationIndex,:));
                
                candidateBasis = orth(subsamples);
                distance = point_to_space_distance(currentInliers,candidateBasis);
                
                inlierIndices = find(distance<=boundaryThreshold);
                outlierIndices = find(distance>boundaryThreshold);
                
                if length(inlierIndices)>maxDegeneracyConsensus
                    maxDegeneracyConsensus = length(inlierIndices);
                    maxDegeneracyOutlierIndices = outlierIndices;
                end
            end
            
            % Update currentInliers
            if  length(maxDegeneracyOutlierIndices)/maxSubspaceConsensus<degeneracyThreshold
                degeneracyExist = true;
                break;
            else
                currentInliers = currentInliers(:,maxDegeneracyOutlierIndices);
                currentInlierCount = length(maxDegeneracyOutlierIndices);
            end
        end
        
        if degeneracyExist
            % Trim out this degenerate sample set
            currentX=currentX(:,maxOutlierIndices);
            currentXCount = length(maxOutlierIndices);
            
            % Re-estimate a subspace model
            maxSubspaceConsensus = 0;
            if (currentXCount<currentSubspaceDimension)
                error('RANSAC failed to segment a generally distributed subspace model.');
            end
            adjustedIterationNumber = nchoosek(currentXCount,currentSubspaceDimension);
            if iterationNumber>adjustedIterationNumber
                newSamplingIndices = generate_subsets(currentXCount,currentSubspaceDimension);
            else
                adjustedIterationNumber = iterationNumber;
                newSamplingIndices = [];
                for iterationIndex=1:iterationNumber
                    newSamplingIndices(iterationIndex,:) = generate_random_subset(currentXCount,currentSubspaceDimension);
                end
            end
            for iterationIndex=1:adjustedIterationNumber
                % Step 1: generate a minimal sample set
                subsamples=currentX(:,newSamplingIndices(iterationIndex,:));

                % Step 2: Calculate an orthogonal basis
                candidateBasis = orth(subsamples);

                % Step 3: Calculate the consensus
                distance = point_to_space_distance(currentX,candidateBasis);
                outlierIndices = find(distance>boundaryThreshold);
                inlierIndices = find(distance<=boundaryThreshold);

                if length(inlierIndices)>maxSubspaceConsensus
                    maxSubspaceConsensus = length(inlierIndices);
                    maxOutlierIndices = outlierIndices;
                    maxInlierIndices = inlierIndices;
                end
            end
            
            possibleIntersectionNumber = possibleIntersectionNumber - 1;
        else
            currentX = currentX(:,maxInlierIndices);
        end
    end
    
    % Compute basis
    [U,S,V]=svds(currentX, currentSubspaceDimension);
    newSubspaceBases{subspaceIndex} = U;
    
    % Update leftX
    distance = point_to_space_distance(leftX,U);
    leftX = leftX(:,find(distance>boundaryThreshold));
    leftXCount = size(leftX,2);
end

% Step 4. Recalculate the subspace structures


% Re-map the sorted subspaces
inverseIndex(sortedIndex) = 1:subspaceCount;

% Determine the membership of all samples based on the assumed structure
distance = [];
for subspaceIndex=1:subspaceCount
    subspaceBases{subspaceIndex} = newSubspaceBases{inverseIndex(subspaceIndex)};
    distance(subspaceIndex,:) = point_to_space_distance(X,subspaceBases{subspaceIndex});
end

% Assign subspace label
for sampleIndex=1:sampleCount
    [minDistance, distancePointer]=min(distance(:,sampleIndex));
    if minDistance>boundaryThreshold
        % Outlier
        sampleLabels(sampleIndex) = -1;
    else
        sampleLabels(sampleIndex) = distancePointer;
    end
end

% Post-processing
if postOptimization
    inlierIndex = find(sampleLabels~=-1);
    [sampleLabels(inlierIndex), subspaceBases, returnStatus]=Ksubspaces(X(:,inlierIndex), subspaceDimensions, 'initialBases',subspaceBases);
end
if dataNormalization
    X = normalize_variance(X,subspaceVariances);
    for subspaceIndex=1:subspaceCount
        subspaceX=X(:,find(sampleLabels==subspaceIndex));
        [U,S,V]=svds(subspaceX, subspaceDimensions(subspaceIndex));
        subspaceBases{subspaceIndex} = U;
    end
end